﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Core.Views;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.MainForm;
using HeuristicLab.Visualization;

namespace HeuristicLab.EvolutionaryTracking.Views {
  [View("SymbolicExpressionTreeGenealogyGraph")]
  [Content(typeof(SymbolicExpressionTreeGenealogyGraph), IsDefaultView = true)]
  public sealed partial class GenealogyGraphView : ItemView {
    private VisualSymbolicExpressionTreeNode _selectedVisualSymbolicExpressionTreeNode;

    public new SymbolicExpressionTreeGenealogyGraph Content {
      get { return (SymbolicExpressionTreeGenealogyGraph)base.Content; }
      set { base.Content = value; }
    }

    public GenealogyGraphView()
      : base() {
      InitializeComponent();
      similarityModeSelector.SelectedIndex = 0; // set default similarity mode to "exact"
    }

    protected override void DeregisterContentEvents() {
      // TODO: Deregister your event handlers here
      genealogyGraphChart.GenealogyGraphNodeClicked -= graphChart_GenealogyGraphNodeClicked;
      symbolicExpressionTreeChart.SymbolicExpressionTreeNodeClicked -= treeChart_SymbolicExpressionTreeNodeClicked;
      base.DeregisterContentEvents();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      // TODO: Register your event handlers here
      genealogyGraphChart.GenealogyGraphNodeClicked += graphChart_GenealogyGraphNodeClicked;
      symbolicExpressionTreeChart.SymbolicExpressionTreeNodeClicked += treeChart_SymbolicExpressionTreeNodeClicked;
    }

    #region Event Handlers (Content)
    // TODO: Put event handlers of the content here
    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content == null)
        genealogyGraphChart.Graph = null;
      else {
        genealogyGraphChart.Graph = Content;
        var best = Content.Values.OrderByDescending(x => x.Quality).First();
        symbolicExpressionTreeChart.Tree = (ISymbolicExpressionTree)best.Data;
      }
    }

    private void Content_GraphChanged(object sender, EventArgs e) {
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      genealogyGraphChart.Enabled = Content != null;
    }
    #endregion

    #region Event Handlers (child controls)

    // TODO: Put event handlers of child controls here.
    private void graphChart_GenealogyGraphNodeClicked(object sender, MouseEventArgs e) {
      // the type hierarchy goes like this: VisualGenealogyGraphNode --> GenealogyGraphNode --> symbolicExpressionTree
      var visualGenealogyGraphNode = (VisualGenealogyGraphNode)sender;
      var genealogyGraphNode = (GenealogyGraphNode)visualGenealogyGraphNode.Data;
      symbolicExpressionTreeChart.Tree = (ISymbolicExpressionTree)genealogyGraphNode.Data;
      if (_selectedVisualSymbolicExpressionTreeNode != null) {
        var nodes = symbolicExpressionTreeChart.Tree.IterateNodesBreadth() as List<ISymbolicExpressionTreeNode>;
        var fragments = _selectedVisualSymbolicExpressionTreeNode.SymbolicExpressionTreeNode.IterateNodesBreadth() as List<ISymbolicExpressionTreeNode>;
        int index = SymbolicExpressionTreeMatching.FindMatch(nodes, fragments, similarityModeSelector.SelectedIndex);
        if (index != -1) {
          _selectedVisualSymbolicExpressionTreeNode = symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(nodes[index]);
          var subtree = _selectedVisualSymbolicExpressionTreeNode.SymbolicExpressionTreeNode;
          foreach (var visualNode in subtree.IterateNodesBreadth().Select(node => symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node))) {
            visualNode.FillColor = Color.LightBlue;
          }
          symbolicExpressionTreeChart.Repaint();
        }
      }
      var tree = symbolicExpressionTreeChart.Tree;
      var graphNode = Content.GetNode(tree);
      if (graphNode.InEdges != null) {
        var arc = graphNode.InEdges.Find(a => a.Data != null);
        if (arc != null) {
          var fragment = arc.Data as ISymbolicExpressionTreeNode;
          foreach (var node in fragment.IterateNodesBreadth()) {
            var visualNode = symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node);
            visualNode.FillColor = Color.LightGreen;
          }
          symbolicExpressionTreeChart.Repaint();
        }
      }
    }

    private void moveModeButton_CheckedChanged(object sender, EventArgs e) {
      var btn = (RadioButton)sender;
      if (btn.Checked) { genealogyGraphChart.Chart.Mode = ChartMode.Move; }
    }

    private void zoomModeButton_CheckedChanged(object sender, EventArgs e) {
      var btn = (RadioButton)sender;
      if (btn.Checked) { genealogyGraphChart.Chart.Mode = ChartMode.Zoom; }
    }

    private void selectModeButton_CheckedChanged(object sender, EventArgs e) {
      var btn = (RadioButton)sender;
      if (btn.Checked) { genealogyGraphChart.Chart.Mode = ChartMode.Select; }
    }

    private void treeChart_SymbolicExpressionTreeNodeClicked(object sender, MouseEventArgs e) {
      _selectedVisualSymbolicExpressionTreeNode = (VisualSymbolicExpressionTreeNode)sender;
      // find out which individuals in the genealogy graph contain this specific subtree
      var treeNode = _selectedVisualSymbolicExpressionTreeNode.SymbolicExpressionTreeNode;
      Color[] colors = { Color.LightSkyBlue, Color.PaleGreen, Color.Tan };

      genealogyGraphChart.ClearAllNodes(); // clear node colors
      // color each graph node according to the degree to which it matches the selected tree fragment
      foreach (var i in Enum.GetValues(typeof(SymbolicExpressionTreeMatching.SimilarityLevel)).Cast<int>().Reverse()) {
        var owners = genealogyGraphChart.Graph.TraceFragment(treeNode, i).ToList();
        if (owners.Any()) genealogyGraphChart.HighlightNodes(owners, colors[i]); // highlight matching individuals from the genealogy
      }

      // highlight subtree nodes in the tree chart
      foreach (var visualNode in symbolicExpressionTreeChart.Tree.IterateNodesPostfix().Select(node => symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node)))
        visualNode.FillColor = Color.Transparent;

      foreach (var visualNode in treeNode.IterateNodesPostfix().Select(node => symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node)))
        visualNode.FillColor = Color.LightBlue;

      // refresh the tree chart
      symbolicExpressionTreeChart.Repaint();
    }

    private void similarityModeSelector_SelectedIndexChanged(object sender, EventArgs e) {
      if (_selectedVisualSymbolicExpressionTreeNode == null) return;
      var treeNode = _selectedVisualSymbolicExpressionTreeNode.SymbolicExpressionTreeNode;
      var owners = genealogyGraphChart.Graph.TraceFragment(treeNode, similarityModeSelector.SelectedIndex).ToList();
      if (owners.Any()) {
        genealogyGraphChart.ClearAllNodes(); // clear the fill color of all nodes
        genealogyGraphChart.HighlightNodes(owners, Color.LightSkyBlue); // highlight matching individuals from the genealogy
      }
      // highlight subtree nodes in the tree chart
      foreach (var visualNode in symbolicExpressionTreeChart.Tree.IterateNodesPostfix().Select(node => symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node)))
        visualNode.FillColor = Color.Transparent;
      foreach (var visualNode in treeNode.IterateNodesPostfix().Select(node => symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node)))
        visualNode.FillColor = Color.LightBlue;
    }
    #endregion
  }
}
